from .msv2_to_msv4_meta import column_description_casacore_to_msv4_measure
from .subtables import subt_rename_ids
from ._tables.read import read_generic_table


def create_field_info(xds, in_file, field_id):
    field_xds = read_generic_table(
        in_file,
        "FIELD",
        rename_ids=subt_rename_ids["FIELD"],
    ).sel(field_id=field_id)
    # https://stackoverflow.com/questions/53195684/how-to-navigate-a-dict-by-list-of-keys

    field_column_description = field_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]
    # ['DELAY_DIR', 'PHASE_DIR', 'REFERENCE_DIR', 'CODE', 'FLAG_ROW', 'NAME', 'NUM_POLY', 'SOURCE_ID', 'TIME']

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["REFERENCE_DIR"],
        ref_code=getattr(field_xds.get("refdir_ref"), "data", None),
    )
    delay_dir = {
        "dims": "",
        "data": list(field_xds["delay_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["PHASE_DIR"],
        ref_code=getattr(field_xds.get("phasedir_ref"), "data", None),
    )
    phase_dir = {
        "dims": "",
        "data": list(field_xds["phase_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["DELAY_DIR"],
        ref_code=getattr(field_xds.get("delaydir_ref"), "data", None),
    )
    reference_dir = {
        "dims": "",
        "data": list(field_xds["delay_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    field_info = {
        "name": str(field_xds["name"].data),
        "code": str(field_xds["code"].data),
        "delay_direction": delay_dir,
        "phase_direction": phase_dir,
        "reference_direction": reference_dir,
        "field_id": field_id,
    }
    xds.attrs["field_info"] = field_info

    return xds



import numcodecs
import time
from .._zarr.encoding import add_encoding
from typing import Dict, Union

import numpy as np
import xarray as xr

from .msv4_infos import create_field_info
from .msv4_sub_xdss import create_ant_xds, create_pointing_xds, create_weather_xds
from .msv2_to_msv4_meta import (
    column_description_casacore_to_msv4_measure,
    create_attribute_metadata,
    col_to_data_variable_names,
    col_dims,
)
from .partition_queries import create_taql_query_and_file_name
from .subtables import subt_rename_ids
from ._tables.table_query import open_table_ro, open_query
from ._tables.read import (
    convert_casacore_time,
    extract_table_attributes,
    read_col_conversion,
    read_generic_table,
)
from ._tables.read_main_table import get_baselines, get_baseline_indices, get_utimes_tol
from .._utils.stokes_types import stokes_types
from xradio.vis._vis_utils._ms.optimised_functions import unique_1d


def check_if_consistent(col, col_name):
    """_summary_

    Parameters
    ----------
    col : _type_
        _description_
    col_name : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """

    col_unique = unique_1d(col)
    assert len(col_unique) == 1, col_name + " is not consistent."
    return col_unique[0]


# TODO: if the didxs are not used in read_col_conversion, remove didxs from here (and convert_and_write_partition)
def calc_indx_for_row_split(tb_tool, taql_where):
    baselines = get_baselines(tb_tool)
    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]

    freq_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]
    utimes, tol = get_utimes_tol(tb_tool, taql_where)

    tidxs = np.searchsorted(utimes, tb_tool.getcol("TIME"))

    ts_ant1, ts_ant2 = (
        tb_tool.getcol("ANTENNA1"),
        tb_tool.getcol("ANTENNA2"),
    )

    ts_bases = np.column_stack((ts_ant1, ts_ant2))
    bidxs = get_baseline_indices(baselines, ts_bases)

    # some antenna 2"s will be out of bounds for this chunk, store rows that are in bounds

    didxs = np.where((bidxs >= 0) & (bidxs < len(baselines)))[0]

    baseline_ant1_id = baselines[:, 0]
    baseline_ant2_id = baselines[:, 1]

    return (
        tidxs,
        bidxs,
        didxs,
        baseline_ant1_id,
        baseline_ant2_id,
        convert_casacore_time(utimes, False),
    )


def create_coordinates(
    xds, in_file, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
):
    coords = {
        "time": utime,
        "baseline_antenna1_id": ("baseline_id", baseline_ant1_id),
        "baseline_antenna2_id": ("baseline_id", baseline_ant2_id),
        "uvw_label": ["u", "v", "w"],
        "baseline_id": np.arange(len(baseline_ant1_id)),
    }

    ddi_xds = read_generic_table(in_file, "DATA_DESCRIPTION").sel(row=ddi)
    pol_setup_id = ddi_xds.polarization_id.values
    spw_id = ddi_xds.spectral_window_id.values

    spw_xds = read_generic_table(
        in_file,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    ).sel(spectral_window_id=spw_id)
    coords["frequency"] = spw_xds["chan_freq"].data[
        ~(np.isnan(spw_xds["chan_freq"].data))
    ]

    pol_xds = read_generic_table(
        in_file,
        "POLARIZATION",
        rename_ids=subt_rename_ids["POLARIZATION"],
    )
    num_corr = int(pol_xds["num_corr"][pol_setup_id].values)
    coords["polarization"] = np.vectorize(stokes_types.get)(
        pol_xds["corr_type"][pol_setup_id, :num_corr].values
    )

    xds = xds.assign_coords(coords)

    ###### Create Frequency Coordinate ######
    freq_column_description = spw_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_FREQ"], ref_code=spw_xds["meas_freq_ref"].data
    )
    xds.frequency.attrs.update(msv4_measure)

    xds.frequency.attrs["spectral_window_name"] = str(spw_xds.name.values)
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["REF_FREQUENCY"], ref_code=spw_xds["meas_freq_ref"].data
    )
    xds.frequency.attrs["reference_frequency"] = {
        "dims": "",
        "data": float(spw_xds.ref_frequency.values),
        "attrs": msv4_measure,
    }
    xds.frequency.attrs["spw_id"] = spw_id

    # xds.frequency.attrs["effective_channel_width"] = "EFFECTIVE_CHANNEL_WIDTH"
    # Add if doppler table is present
    # xds.frequency.attrs["doppler_velocity"] =
    # xds.frequency.attrs["doppler_type"] =

    unique_chan_width = unique_1d(
        spw_xds.chan_width.data[np.logical_not(np.isnan(spw_xds.chan_width.data))]
    )
    # assert len(unique_chan_width) == 1, "Channel width varies for spw."
    # xds.frequency.attrs["channel_width"] = spw_xds.chan_width.data[
    #    ~(np.isnan(spw_xds.chan_width.data))
    # ]  # unique_chan_width[0]
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_WIDTH"], ref_code=spw_xds["meas_freq_ref"].data
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["Hz"]
    xds.frequency.attrs["channel_width"] = {
        "dims": "",
        "data": np.abs(unique_chan_width[0]),
        "attrs": msv4_measure,
    }

    ###### Create Time Coordinate ######
    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["TIME"]
    )
    xds.time.attrs.update(msv4_measure)

    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["INTERVAL"]
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["s"]
    xds.time.attrs["integration_time"] = {
        "dims": "",
        "data": interval,
        "attrs": msv4_measure,
    }
    xds.time.attrs["effective_integration_time"] = "EFFECTIVE_INTEGRATION_TIME"
    return xds


def create_data_variables(
    in_file, xds, tb_tool, time_baseline_shape, tidxs, bidxs, didxs
):
    # Create Data Variables
    col_names = tb_tool.colnames()

    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    for col in col_names:
        if col in col_to_data_variable_names:
            if (col == "WEIGHT") and ("WEIGHT_SPECTRUM" in col_names):
                continue
            try:
                start = time.time()
                if col == "WEIGHT":
                    xds[col_to_data_variable_names[col]] = xr.DataArray(
                        np.tile(
                            read_col_conversion(
                                tb_tool,
                                col,
                                time_baseline_shape,
                                tidxs,
                                bidxs,
                            )[:, :, None, :],
                            (1, 1, xds.sizes["frequency"], 1),
                        ),
                        dims=col_dims[col],
                    )

                else:
                    xds[col_to_data_variable_names[col]] = xr.DataArray(
                        read_col_conversion(
                            tb_tool,
                            col,
                            time_baseline_shape,
                            tidxs,
                            bidxs,
                        ),
                        dims=col_dims[col],
                    )
                    # logger.info("Time to read column " + str(col) + " : " + str(time.time()-start))
            except:
                # logger.debug("Could not load column",col)
                continue

            xds[col_to_data_variable_names[col]].attrs.update(
                create_attribute_metadata(col, main_column_descriptions)
            )


def convert_and_write_partition(
    in_file: str,
    out_file: str,
    intent: str,
    ddi: int = 0,
    state_ids=None,
    field_id: int = None,
    ignore_msv2_cols: Union[list, None] = None,
    main_chunksize: Union[Dict, None] = None,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend="zarr",
    overwrite: bool = False,
):
    """_summary_

    Parameters
    ----------
    in_file : str
        _description_
    out_file : str
        _description_
    intent : str
        _description_
    ddi : int, optional
        _description_, by default 0
    state_ids : _type_, optional
        _description_, by default None
    field_id : int, optional
        _description_, by default None
    ignore_msv2_cols : Union[list, None], optional
        _description_, by default None
    main_chunksize : Union[Dict, None], optional
        _description_, by default None
    compressor : numcodecs.abc.Codec, optional
        _description_, by default numcodecs.Zstd(level=2)
    storage_backend : str, optional
        _description_, by default "zarr"
    overwrite : bool, optional
        _description_, by default False

    Returns
    -------
    _type_
        _description_
    """
    if ignore_msv2_cols is None:
        ignore_msv2_cols = []

    taql_where, file_name = create_taql_query_and_file_name(
        out_file, intent, state_ids, field_id, ddi
    )

    start_with = time.time()
    with open_table_ro(in_file) as mtable:
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:
            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            (
                tidxs,
                bidxs,
                didxs,
                baseline_ant1_id,
                baseline_ant2_id,
                utime,
            ) = calc_indx_for_row_split(tb_tool, taql_where)
            time_baseline_shape = (len(utime), len(baseline_ant1_id))
            # logger.debug("Calc indx for row split "+ str(time.time()-start))

            xds = xr.Dataset()
            # interval = check_if_consistent(tb_tool.getcol("INTERVAL"), "INTERVAL")
            interval = tb_tool.getcol("INTERVAL")

            interval_unique = unique_1d(interval)
            if len(interval_unique) > 1:
                print(
                    "Integration time (interval) not consitent in partition, using median."
                )
                interval = np.median(interval)
            else:
                interval = interval_unique[0]

            xds = create_coordinates(
                xds, in_file, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
            )

            create_data_variables(
                in_file, xds, tb_tool, time_baseline_shape, tidxs, bidxs, didxs
            )

            # Create field_info
            field_id = check_if_consistent(tb_tool.getcol("FIELD_ID"), "FIELD_ID")
            xds = create_field_info(xds, in_file, field_id)

            # Create ant_xds
            ant_xds = create_ant_xds(in_file)

            # Create weather_xds
            weather_xds = create_weather_xds(in_file)

            pointing_xds = create_pointing_xds(in_file)

            # To do: add other _info and _xds creation functions.

            # Fix UVW frame
            # From CASA fixvis docs: clean and the im tool ignore the reference frame claimed by the UVW column (it is often mislabelled as ITRF when it is really FK5 (J2000)) and instead assume the (u, v, w)s are in the same frame as the phase tracking center. calcuvw does not yet force the UVW column and field centers to use the same reference frame! Blank = use the phase tracking frame of vis.
            xds.UVW.attrs["frame"] = xds.attrs["field_info"]["phase_direction"][
                "attrs"
            ]["frame"]

            xds.attrs["intent"] = intent
            xds.attrs["ddi"] = ddi

            # Time and frequency should always be increasing
            if len(xds.frequency) > 1 and xds.frequency[1] - xds.frequency[0] < 0:
                xds = xds.sel(frequency=slice(None, None, -1))

            if len(xds.time) > 1 and xds.time[1] - xds.time[0] < 0:
                xds = xds.sel(time=slice(None, None, -1))

            # Add data_groups
            xds.attrs["data_groups"] = {
                "base": {
                    "visibility": "VISIBILITY",
                    "flag": "FLAG",
                    "weight": "WEIGHT",
                    "uvw": "UVW",
                }
            }

            if overwrite:
                mode = "w"
            else:
                mode = "w-"

            add_encoding(xds, compressor=compressor, chunks=main_chunksize)

            if storage_backend == "zarr":
                xds.to_zarr(store=file_name + "/MAIN", mode=mode)
                ant_xds.to_zarr(store=file_name + "/ANTENNA", mode=mode)
                pointing_xds.to_zarr(store=file_name + "/POINTING", mode=mode)
                if weather_xds:
                    weather_xds.to_zarr(store=file_name + "/WEATHER", mode=mode)
            elif storage_backend == "netcdf":
                # xds.to_netcdf(path=file_name+"/MAIN", mode=mode) #Does not work
                raise

    # logger.info("Saved ms_v4 " + file_name + " in " + str(time.time() - start_with) + "s")



import os

import xarray as xr

from ._processing_set import processing_set


def read_processing_set(ps_name, intents=None, fields=None):
    items = os.listdir(ps_name)
    ms_xds = xr.Dataset()
    ps = processing_set()
    for i in items:
        if "ddi" in i:
            xds = xr.open_zarr(ps_name + "/" + i + "/MAIN")

            if (intents is None) or (xds.attrs["intent"] in intents):
                if (fields is None) or (xds.attrs["field_info"]["name"] in fields):
                    ps[i] = xds
                    sub_xds = {
                        "antenna_xds": "ANTENNA",
                        "pointing_xds": "POINTING",
                    }
                    for sub_xds_key, sub_xds_name in sub_xds.items():
                        ps[i].attrs[sub_xds_key] = xr.open_zarr(
                            ps_name + "/" + i + "/" + sub_xds_name
                        )

                    optional_sub_xds = {
                        "weather_xds": "WEATHER",
                    }
                    for sub_xds_key, sub_xds_name in optional_sub_xds.items():
                        sub_xds_path = ps_name + "/" + i + "/" + sub_xds_name
                        if os.path.isdir(sub_xds_path):
                            ps[i].attrs[sub_xds_key] = xr.open_zarr(sub_xds_path)

    return ps



"""Contains optimised functions to be used within other modules."""

import numpy as np
import pandas as pd


def unique_1d(array: np.ndarray) -> np.ndarray:
    """Optimised version of np.unique for 1D arrays.

    Args:
        array (np.ndarray): a 1D array of values.

    Returns:
        np.ndarray: a sorted array of unique values.
    """
    return np.sort(pd.unique(array))


def pairing_function(antenna_pairs: np.ndarray) -> np.ndarray:
    """Pairing function to convert each array pair to a single value.

    This custom pairing function will only work if the maximum value is less
    than 2**20 and less than 2,048 if using signed 32-bit integers.

    Args:
        antenna_pairs (np.ndarray): a 2D array containing antenna 1 and antenna
        2 ids, which forms a baseline.

    Returns:
        np.ndarray: a 1D array of the paired values.
    """
    return antenna_pairs[:, 0] * 2**20 + antenna_pairs[:, 1]


def inverse_pairing_function(paired_array: np.ndarray) -> np.ndarray:
    """Inverse pairing function to convert each paired value to an antenna pair.

    This inverse pairing function is the inverse of the custom pairing function.

    Args:
        paired_array (np.ndarray): a 1D array of the paired values.

    Returns:
        np.ndarray: a 2D array containing antenna 1 and antenna 2 ids.
    """
    return np.column_stack(np.divmod(paired_array, 2**20))



from importlib_metadata import version
import graphviper.utils.logger as logger, multiprocessing, psutil
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from .cds import CASAVisSet
from .stokes_types import stokes_types


def make_coords(
    xds: xr.Dataset, ddi: int, subtables: Tuple[xr.Dataset, ...]
) -> Dict[str, np.ndarray]:
    """Make the coords to be added to a partition or chunk (besides
    the time, baseline) basic structure

    Grabs:
    - channel (center) frequency values from the spw subtable
    - pol idxs from the pol+ddi subtables -> pol names via the stokes_types
    - antenna IDs from antenna subtable

    :param: sub-xds as (ant_xds, ddi_xds, spw_xds, pol_xds)
    """
    ant_xds, ddi_xds, spw_xds, pol_xds = subtables
    freq = spw_xds.chan_freq.values[
        ddi_xds.spectral_window_id.values[ddi], : xds.freq.shape[0]
    ]
    pol_ids = pol_xds.corr_type.values[
        ddi_xds.polarization_id.values[ddi], : xds.pol.shape[0]
    ]
    pol_names = np.vectorize(stokes_types.get)(pol_ids)
    ant_id = ant_xds.antenna_id.values
    coords = {
        "freq": freq,
        "pol": pol_names,
        "antenna_id": ant_id
        # These will be metainfo in partitions
        # "spw_id": [ddi_xds["spectral_window_id"].values[ddi]],
        # "pol_id": [ddi_xds["polarization_id"].values[ddi]],
    }
    return coords


def vis_xds_packager_cds(
    subtables: List[Tuple[str, xr.Dataset]],
    partitions: Dict[Any, xr.Dataset],
    descr_add: str = "",
):
    """Takes a a list of subtable xds datasets and a dictionary of data
    partition xds datasets and and packages them as a CASA vis dataset
    (cds)

    :param partitions: data partiions as xds datasets
    :param subtables: subtables as xds datasets
    :param descr_add: substring to add to the short descr string of the cds
    :return: A "cds" - container for the metainfo subtables and data partitions
    """
    vers = version("xradio")

    return CASAVisSet(
        subtables,
        partitions,
        f"CASA vis set produced by xradio {vers}/{descr_add}",
    )


def vis_xds_packager_mxds(
    partitions: Dict[Any, xr.Dataset],
    subtables: List[Tuple[str, xr.Dataset]],
    add_global_coords: bool = True,
) -> xr.Dataset:
    """Takes a dictionary of data partition xds datasets and a list of
    subtable xds datasets and packages them as a dataset of datasets
    (mxds)

    :param partitions: data partiions as xds datasets
    :param subtables: subtables as xds datasets
    :add_global_coords: whether to add coords to the output mxds
    :return: A "mxds" - xr.dataset of datasets
    """
    mxds = xr.Dataset(attrs={"metainfo": subtables, "partitions": partitions})

    if add_global_coords:
        mxds = mxds.assign_coords(make_global_coords(mxds))

    return mxds


def make_global_coords(mxds: xr.Dataset):
    coords = {}
    metainfo = mxds.attrs["metainfo"]
    if "antenna" in metainfo:
        coords["antenna_ids"] = metainfo["antenna"].antenna_id.values
        coords["antennas"] = xr.DataArray(
            metainfo["antenna"].name.values, dims=["antenna_ids"]
        )
    if "field" in metainfo:
        coords["field_ids"] = metainfo["field"].field_id.values
        coords["fields"] = xr.DataArray(
            metainfo["field"].name.values, dims=["field_ids"]
        )
    if "feed" in mxds.attrs:
        coords["feed_ids"] = metainfo["feed"].feed_id.values
    if "observation" in metainfo:
        coords["observation_ids"] = metainfo["observation"].observation_id.values
        coords["observations"] = xr.DataArray(
            metainfo["observation"].project.values, dims=["observation_ids"]
        )
    if "polarization" in metainfo:
        coords["polarization_ids"] = metainfo["polarization"].pol_setup_id.values
    if "source" in metainfo:
        coords["source_ids"] = metainfo["source"].source_id.values
        coords["sources"] = xr.DataArray(
            metainfo["source"].name.values, dims=["source_ids"]
        )
    if "spectral_window" in metainfo:
        coords["spw_ids"] = metainfo["spectral_window"].spw_id.values
    if "state" in metainfo:
        coords["state_ids"] = metainfo["state"].state_id.values

    return coords


def expand_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    expand single (row) dimension of xds to (time, baseline)
    """
    assert "baseline" not in xds.coords

    txds = xds.copy()
    unique_baselines, baselines = np.unique(
        [txds.antenna1.values, txds.antenna2.values], axis=1, return_inverse=True
    )
    txds["baseline"] = xr.DataArray(baselines.astype("int32"), dims=["row"])

    try:
        txds = (
            txds.set_index(row=["time", "baseline"])
            .unstack("row")
            .transpose("time", "baseline", ...)
        )

        # unstack changes type to float when it needs to introduce NaNs, so
        # we need to reset to the proper type. Avoid if possible, as the
        # astype are costly
        for dv in txds.data_vars:
            if txds[dv].dtype != xds[dv].dtype:
                txds[dv] = txds[dv].astype(xds[dv].dtype)
    except Exception as exc:
        print(
            f"WARNING: Cannot expand rows to (time, baseline), "
            f"possibly duplicate values in (time, baseline). Exception: {exc}"
        )
        txds = xds.copy()

    return txds


def flatten_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    flatten (time, baseline) dimensions of xds back to single dimension (row)
    """
    nan_int = np.array([np.nan]).astype("int32")[0]
    txds = xds.copy()

    # flatten the time x baseline dimensions of main table
    if ("time" in xds.sizes) and ("baseline" in xds.sizes):
        txds = xds.stack({"row": ("time", "baseline")}).transpose("row", ...)
        txds = txds.where(
            (txds.state_id != nan_int) & (txds.field_id != nan_int), drop=True
        )  # .unify_chunks()
        for dv in list(xds.data_vars):
            txds[dv] = txds[dv].astype(xds[dv].dtype)

    return txds


####################################
# xautomatically compute best data chunking
def optimal_chunking(
    ndim: Union[int, None] = None,
    didxs: Union[Tuple[int], List[int], None] = None,
    chunk_size: str = "auto",
    data_shape: Union[tuple, None] = None,
) -> tuple:
    """
    Determine the optimal chunk shape for reading an MS or Image based
    on machine resources and intended operations

    :param ndim: number of dimensions to chunk. An MS is 3, an
    expanded MS is 4. An image could be anywhere from 2 to 5. Not
    needed if data_shape is given.
    :param didxs: dimension indices over which subsequent operations
    will be performed. Values should be less than ndim. Tries to
    reduce inter-process communication of data contents. Needs to
    know the shape to do this well. Default None balances chunk size
    across all dimensions.
    :param chunk_size: target chunk size ('large', 'small', 'auto').
    Default 'auto' tries to guess by looking at CPU core count and
    available memory.
    :param data_shape: shape of the total MS DDI or Image data. Helps
    to know. Default None does not optimize based on shape

    :return: optimal chunking for reading the ms (row, chan, pol)
    """
    assert (ndim is not None) or (
        data_shape is not None
    ), "either ndim or data_shape must be given"
    assert chunk_size in ["large", "small", "auto"], "invalid chunk_size parameter"
    if ndim is None:
        ndim = len(data_shape)

    opt_dims = (
        didxs if (didxs is not None) and (len(didxs) > 0) else np.arange(ndim)
    )  # maximize these dim chunk sizes
    nonopt_dims = np.setdiff1d(np.arange(ndim), opt_dims)  # at the expense of these

    max_chunk_sizes = (
        data_shape
        if data_shape is not None
        else [dd for ii, dd in enumerate([10000, 10000, 10000, 4, 10]) if ii < ndim]
    )
    min_chunk_sizes = (
        np.ceil(np.array(data_shape) / 80).astype(int)
        if data_shape is not None
        else [1000, 1, 1]
        if ndim == 3
        else [dd for ii, dd in enumerate([10, 10, 1, 1, 1]) if ii < ndim]
    )
    target_size = (
        175 * 1024**2 / 8
    )  # ~175 MB chunk worst case with 8-byte DATA column
    bytes_per_core = int(
        round(
            ((psutil.virtual_memory().available * 0.10) / multiprocessing.cpu_count())
        )
    )
    if data_shape is not None:
        bytes_per_core = min(
            bytes_per_core, np.prod(data_shape) * 8 / 2
        )  # ensure at least two chunks
    if chunk_size == "large":
        target_size = target_size * 6  # ~1 GB
    if chunk_size == "auto":
        target_size = max(min(target_size * 6, bytes_per_core / 8), target_size)

    # start by setting the optimized dims to their max size and non-optimized dims to their min size
    chunks = np.zeros((ndim), dtype="int")
    chunks[opt_dims] = np.array(max_chunk_sizes)[opt_dims]
    chunks[nonopt_dims] = np.array(min_chunk_sizes)[nonopt_dims]

    # iteratively walk towards an optimal chunk size
    # iteration is needed because rounding to nearest integer index can make a big different (2x) in chunk size
    # for small dimensions like pol
    for ii in range(10):
        # if the resulting size is too big, reduce the sizes of the optimized dimensions
        if (np.prod(chunks) > target_size) and (len(opt_dims) > 0):
            chunks[opt_dims] = np.round(
                chunks[opt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(opt_dims))
            )
        # else if the resulting size is too small, increase the sizes of the non-optimized dimensions
        elif (np.prod(chunks) < target_size) and (len(nonopt_dims) > 0):
            chunks[nonopt_dims] = np.round(
                chunks[nonopt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(nonopt_dims))
            )
        chunks = np.min((chunks, max_chunk_sizes), axis=0)
        chunks = np.max((chunks, min_chunk_sizes), axis=0)

    return tuple(chunks)


def calc_optimal_ms_chunk_shape(
    memory_available_in_bytes, shape, element_size_in_bytes, column_name
) -> int:
    """
    Calculates the max number of rows (1st dim in shape) of a variable
    that can be fit in the memory for a thread.
    """
    factor = 0.8  # Account for memory used by other objects in thread.
    # total_mem = np.prod(shape)*element_size_in_bytes
    single_row_mem = np.prod(shape[1:]) * element_size_in_bytes

    try:
        assert single_row_mem < factor * memory_available_in_bytes
    except AssertionError as err:
        logger.exception(
            "Not engough memory in a thread to contain a row of "
            + column_name
            + ". Need at least "
            + str(single_row_mem / factor)
            + " bytes."
        )
        raise err

    rows_chunk_size = int((factor * memory_available_in_bytes) / single_row_mem)

    if rows_chunk_size > shape[0]:
        rows_chunk_size = shape[0]

    logger.debug(
        "Numbers of rows in chunk for " + column_name + ": " + str(rows_chunk_size)
    )

    return rows_chunk_size



import os
from graphviper.utils.logger import setup_logger

if not os.getenv("VIPER_LOGGER_NAME"):
    os.environ["VIPER_LOGGER_NAME"] = "xradio"
    setup_logger(
        logger_name="xradio",
        log_to_term=True,
        log_to_file=False,
        log_file="xradio-logfile",
        log_level="INFO",
    )


from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional, Union
from xradio.schema.bases import AsDataArray, AsDataset
from xradio.schema.typing import Attr, Coord, Coordof, Data, Dataof, Name
import numpy

# Dimensions
Time = Literal["time"]
""" Observation time dimension """
AntennaId = Literal["antenna_id"]
""" Antenna ID dimension """
BaselineId = Literal["baseline_id"]
""" Baseline ID dimension """
Frequency = Literal["frequency"]
""" Frequency dimension """
Polarization = Literal["polarization"]
""" Polarization dimension """
UvwLabel = Literal["uvw_label"]
""" Coordinate dimension of UVW data (typically shape 3 for 'u', 'v', 'w') """
DirectionLabel = Literal["direction_label"]
""" Coordinate dimension of UVW data (typically shape 3 for 'u', 'v', 'w') """
TimePolynomial = Literal["time_polynomial"]
""" For data that is represented as variable in time using Taylor expansion """

# Plain data class models
@dataclass
class SourceInfoDict:
    # TODO
    pass


@dataclass
class FieldInfoDict:
    """
    Field positions for each source.

    Defines a field position on the sky. For interferometers, this is the correlated field position.
    For single dishes, this is the nominal pointing direction.
    """
    name: str
    """Field name."""
    field_id: int
    """Field id"""
    code: str
    """Field code indicating special characteristics of the field; user specified."""
    time_reference: str
    """
    Time reference for the directions and rates. When used in :py:class:`VisibilityXds` should match 
    the scale and format given for ``time`` (see :py:class:`TimeAxis`).
    """
    delay_direction: SkyCoordDict
    
    # TODO
    pass


@dataclass
class QuantityDict:
    # TODO
    pass

@dataclass
class SkyCoordDict:
    # TODO
    pass

@dataclass(frozen=True)
class ObservationInfoDict:
    observer: List[str]
    """List of observer names."""
    project: str
    """Project Code/Project_UID"""
    release_data: str
    """Project release date. This is the date on which the data may become public."""
    execution_block_id: Optional[str]
    execution_block_number: Optional[str]
    execution_block_UID: Optional[str]
    session_reference: Optional[str]
    observing_script: Optional[str]

# Coordinates / Axes
@dataclass(frozen=True)
class TimeAxis(AsDataArray):
    """Data model of time axis"""

    data: Data[Time, float]
    """ Time, expressed in SI seconds since the epoch (see ``scale`` & ``format``). """

    integration_time: Attr[Optional[QuantityDict]] = None
    """ The nominal sampling interval (ms v2). Units of seconds. """
    effective_integration_time: Attr[Optional[QuantityDict]] = None
    """ Name of data array that contains the integration time that includes the effects of missing data. """

    long_name: Attr[str] = "Observation Time"
    """ Long-form name to use for axis"""
    type: Attr[str] = "time"
    """ Coordinate type. Should be ``"time"``. """
    units: Attr[tuple[str]] = ("s",)
    """ Units to associate with axis"""
    scale: Attr[str] = "tai"
    """ Astropy time scales, see :py:class:`astropy.time.Time` """
    format: Attr[str] = "unix"
    """ Astropy format, see :py:class:`astropy.time.Time`. Default seconds from 1970-01-01 00:00:00 UTC """

@dataclass(frozen=True)
class AntennaAxis(AsDataArray):
    data: Data[AntennaId, int]
    """
    Antenna id of an antenna. Maps to ``antenna_id``
    in :py:class:`AntennaXds`.
    """
    long_name: Attr[str] = "Antenna ID"


@dataclass(frozen=True)
class BaselineAxis(AsDataArray):
    """TODO: documentation"""

    data: Data[BaselineId, int]
    """Unique id for each baseline."""
    long_name: Attr[str] = "Baseline ID"


@dataclass(frozen=True)
class BaselineAntennaAxis(AsDataArray):
    data: Data[BaselineId, int]
    """
    Antenna id for an antenna in a baseline. Maps to ``attrs['antenna_xds'].antenna_id``
    in :py:class:`VisibilityXds`
    """
    long_name: Attr[str] = "Baseline Antenna ID"


@dataclass(frozen=True)
class FrequencyAxis(AsDataArray):
    """TODO: documentation"""

    data: Data[Frequency, float]
    """ Time, expressed in SI seconds since the epoch. """
    spectral_window_name: Attr[str]
    """ Name associated with spectral window. """
    frequency_group_name: Attr[str]
    """ Name associated with frequency group - needed for multi-band VLBI fringe-fitting."""
    reference_frequency: Attr[QuantityDict]
    """ A frequency representative of the spectral window, usually the sky
    frequency corresponding to the DC edge of the baseband. Used by the calibration
    system if a ﬁxed scaling frequency is required or in algorithms to identify the
    observing band. """
    channel_width: Attr[QuantityDict]
    """ The nominal channel bandwidth. Same units as data array (see units key). """
    doppler: Optional[Attr[DopplerXds]]
    """ Doppler tracking information """

    type: Attr[str] = "spectral_coord"
    """ Coordinate type. Should be ``"spectral_coord"``. """
    long_name: Attr[str] = "Frequency"
    """ Long-form name to use for axis"""
    units: Attr[tuple[str]] = ("Hz",)
    """ Units to associate with axis"""
    frame: Attr[str] = "icrs"
    """
    Astropy velocity reference frames (see :external:ref:`astropy-spectralcoord`).
    Note that Astropy does not use the name
    'topo' (telescope centric) velocity frame, rather it assumes if no velocity
    frame is given that this is the default.
    """

@dataclass(frozen=True)
class PolarizationAxis(AsDataArray):
    """
    Possible correlations that can be formed from polarised receptors. Possible
    values, taken from `Measures/Stokes.h
    <https://github.com/casacore/casacore/blob/5a8df94738bdc36be27e695d7b14fe949a1cc2df/measures/Measures/Stokes.h>`_:

    * ``I``, ``Q``, ``U``, ``V`` (standard stokes parameters)
    * ``RR``, ``RL``, ``LR``, ``LL`` (circular correlation products)
    * ``XX``, ``XY``, ``YX``, ``YY`` (linear correlation products)
    * ``RX``, ``RY``, ``LX``, ``LY``, ``XR``, ``XL``, ``YR``, ``YL`` (mixed correlation products)
    * ``PP``, ``PQ``, ``QP``, ``QQ`` (general quasi-orthogonal correlation products)
    * ``RCircular``, ``LCircular``, ``Linear`` (single dish polarization types)
    * ``Ptotal`` (polarized intensity: ``sqrt(Q²+U²+V²)``)
    * ``Plinear`` (linearly polarized intensity: ``sqrt(Q²+U²)``)
    * ``PFtotal`` (polarization fraction: ``Ptotal/I``)
    * ``PFlinear`` (linear polarization fraction: ``Plinear/I``)
    * ``Pangle`` (linear polarization angle: ``0.5 arctan(U/Q)`` in radians)

    """

    data: Data[Polarization, str]
    """ Polarization names. """
    type: Attr[str] = "polarization"
    """ Coordinate type. Should be ``"polarization"``. """
    long_name: Attr[str] = "Polarization"
    """ Long-form name to use for axis. Should be ``"Polarization"``"""


@dataclass(frozen=True)
class UvwLabelAxis(AsDataArray):
    """
    Coordinate axis to make up ``("u", "v", "w")`` tuple, see :py:class:`UvwArray`.
    """

    data: Data[UvwLabel, str] = ("u", "v", "w")
    """Should be ``('u','v','w')``, used by :py:class:`UvwArray`"""
    long_name: Attr[str] = "U/V/W label"
    """ Long-form name to use for axis. Should be ``"U/V/W label"``"""

# Data variables
@dataclass(frozen=True)
class VisibilityArray(AsDataArray):
    """TODO: documentation"""

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization],
        numpy.complex64 | numpy.complex128,
    ]
    time: Coordof[TimeAxis]
    baseline_id: Coordof[BaselineAxis]
    frequency: Coordof[FrequencyAxis]
    polarization: Coordof[PolarizationAxis]
    delay_direction: Attr[SkyCoordDict]
    """ 
    Direction of delay center, i.e. what the coorelator originally phased the
    visibilities to.

    For conversion from MSv2, frame refers column
    keywords by default. If frame varies with field, it refers PhaseDir_Ref
    column instead.
    """
    phase_direction: Attr[SkyCoordDict]
    """
    Phase direction of visibilities, i.e. the sky direction from which flux
    would result in real-valued visibilities independent of baseline UVW.
    """
    long_name: Attr[str] = "Visibility values"
    """ Long-form name to use for axis. Should be ``"Visibility values"``"""
    units: Attr[tuple[str]] = ("Jy",)


@dataclass(frozen=True)
class FlagArray(AsDataArray):
    """
    An array of Boolean values with the same shape as `VISIBILITY`,
    representing the cumulative flags applying to this data matrix. Data are
    flagged bad if the ``FLAG`` array element is ``True``.
    """

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization]
        | tuple[Time, BaselineId, Frequency]
        | tuple[Time, BaselineId],
        bool,
    ]
    time: Coordof[TimeAxis]
    baseline_id: Coordof[BaselineAxis]
    frequency: Coordof[FrequencyAxis]
    polarization: Optional[Coordof[PolarizationAxis]] = None
    long_name: Attr[str] = "Visibility flags"


@dataclass(frozen=True)
class WeightArray(AsDataArray):
    """
    The weight for each channel, with the same shape as the associated
    :py:class:`VisibilityArray`, as assigned by the correlator or processor.

    Weight spectrum in ms v2 is renamed weight. Should be calculated as
    1/sigma^2 (sigma rms noise).
    """

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization]
        | tuple[Time, BaselineId, Frequency]
        | tuple[Time, BaselineId],
        numpy.float16 | numpy.float32 | numpy.float64,
    ]
    """Visibility weights"""
    time: Coordof[TimeAxis]
    baseline_id: Coordof[BaselineAxis]
    frequency: Optional[Coordof[FrequencyAxis]] = None
    polarization: Optional[Coordof[PolarizationAxis]] = None
    long_name: Attr[str] = "Visibility weights"


@dataclass(frozen=True)
class UvwArray(AsDataArray):
    """
    Coordinates for the baseline from ``baseline_antenna2_id`` to
    ``baseline_antenna1_id``, i.e. the baseline is equal to the difference
    ``POSITION2 - POSITION1``. The UVW given are for the ``TIME_CENTROID``, and
    correspond in general to the reference type for the
    ``field_info.phase_dir``.

    The baseline direction should be: ``W`` towards source direction; ``V`` in
    plane through source and system's pole; ``U`` in direction of increasing
    longitude coordinate.  So citing
    http://casa.nrao.edu/Memos/CoordConvention.pdf: Consider an XYZ Celestial
    coordinate system centered at the location of the interferometer, with
    :math:`X` towards the East, :math:`Z` towards the NCP and :math:`Y` to
    complete a right-handed system. The UVW coordinate system is then defined
    by the hour-angle and declination of the phase-reference direction such
    that

    #. when the direction of observation is the NCP (`ha=0,dec=90`),
       the UVW coordinates are aligned with XYZ,
    #. V, W and the NCP are always on a Great circle,
    #. when W is on the local meridian, U points East
    #. when the direction of observation is at zero declination, an
       hour-angle of -6 hours makes W point due East.

    This definition also determines the sign of the phase of ``VISIBILITY``.

    """

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization, UvwLabel]
        | tuple[Time, BaselineId, Frequency, UvwLabel]
        | tuple[Time, BaselineId, UvwLabel],
        numpy.float16 | numpy.float32 | numpy.float64,
    ]
    """Baseline coordinates from ``baseline_antenna2_id`` to ``baseline_antenna1_id``"""
    time: Coordof[TimeAxis]
    baseline_id: Coordof[BaselineAxis]
    frequency: Optional[Coordof[FrequencyAxis]] = None
    polarization: Optional[Coordof[PolarizationAxis]] = None
    uvw_label: Coordof[UvwLabelAxis] = ("u", "v", "w")
    long_name: Attr[str] = "Baseline coordinates"
    """ Long-form name to use for axis. Should be ``"Baseline coordinates``"""
    units: Attr[tuple[str]] = ("m",)


@dataclass(frozen=True)
class TimeSamplingArray(AsDataArray):
    """TODO: documentation"""

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization]
        | tuple[Time, BaselineId, Frequency]
        | tuple[Time, BaselineId],
        float,
    ]

    time: Coordof[TimeAxis]
    baseline_id: Coordof[BaselineAxis]
    frequency: Optional[Coordof[FrequencyAxis]] = None
    polarization: Optional[Coordof[PolarizationAxis]] = None

    scale: Attr[str] = "tai"
    """ Astropy time scales, see :py:class:`astropy.time.Time` """
    format: Attr[str] = "unix"
    """ Astropy format, see :py:class:`astropy.time.Time`. Default seconds from 1970-01-01 00:00:00 UTC """

    long_name: Attr[str] = "Time sampling data"
    units: Attr[str] = "s"


@dataclass(frozen=True)
class FreqSamplingArray(AsDataArray):
    """TODO: documentation"""

    data: Data[
        tuple[Time, BaselineId, Frequency, Polarization]
        | tuple[Time, BaselineId, Frequency]
        | tuple[Time, Frequency]
        | tuple[Frequency],
        float,
    ]
    """
    Data about frequency sampling, such as centroid or integration
    time. Concrete function depends on concrete data array within
    :py:class:`VisibilityXds`.
    """
    frequency: Coordof[FrequencyAxis]
    time: Optional[Coordof[TimeAxis]] = None
    baseline_id: Optional[Coordof[BaselineAxis]] = None
    polarization: Optional[Coordof[PolarizationAxis]] = None
    long_name: Attr[str] = "Frequency sampling data"
    units: Attr[str] = "Hz"
    frame: Attr[str] = "icrs"
    """
    Astropy velocity reference frames (see :external:ref:`astropy-spectralcoord`).
    Note that Astropy does not use the name
    'topo' (telescope centric) velocity frame, rather it assumes if no velocity
    frame is given that this is the default.
    """

# Data Sets
@dataclass(frozen=True)
class VisibilityXds(AsDataset):
    """TODO: documentation"""

    # Required Coordinates
    time: Coordof[TimeAxis]
    """
    The time coordinate is the mid-point of the nominal sampling interval, as
    speciﬁed in the ``ms_v4.time.attrs['integration_time']`` (ms v2 interval).
    """
    baseline_id: Coordof[BaselineAxis]
    frequency: Coordof[FrequencyAxis]
    """Center frequencies for each channel."""
    polarization: Coordof[PolarizationAxis]
    """
    Labels for polarization types, e.g. ``['XX','XY','YX','YY']``, ``['RR','RL','LR','LL']``.
    """
    uvw_label: Optional[Coordof[UvwLabelAxis]]

    # Required data variables / arrays
    VISIBILITY: Dataof[VisibilityArray]

    # Required Attributes
    field_info: Attr[FieldInfoDict]
    """Values without any phase rotation. See data array (``VISIBILITY`` and ``UVW``) attribute phase center for phase rotated value. """
    antenna_xds: Attr[AntennaXds]

    # Optional Coordinates
    baseline_antenna1_id: Optional[Coordof[BaselineAntennaAxis]] = None
    """Antenna id for 1st antenna in baseline. Maps to ``attrs['antenna_xds'].antenna_id``"""
    baseline_antenna2_id: Optional[Coordof[BaselineAntennaAxis]] = None
    """Antenna id for 2nd antenna in baseline. Maps to ``attrs['antenna_xds'].antenna_id``"""
    scan_id: Optional[Coord[Time, int]] = None
    """Arbitary scan number to identify data taken in the same logical scan."""

    # Optional data variables / arrays
    """Complex visibilities, either simulated or measured by interferometer."""
    FLAG: Optional[Dataof[FlagArray]] = None
    WEIGHT: Optional[Dataof[WeightArray]] = None
    UVW: Optional[Dataof[UvwArray]] = None
    EFFECTIVE_INTEGRATION_TIME: Optional[Dataof[TimeSamplingArray]] = None
    """
    The integration time, including the effects of missing data, in contrast to
    ``integration_time`` attribute of the ``time`` coordinate,
    see :py:class:`TimeAxis`. (MS v2: ``exposure``).
    """
    TIME_CENTROID: Optional[Dataof[TimeSamplingArray]] = None
    """
    The time centroid of the visibility, includes the effects of missing data
    unlike the ``time`` coordinate, see :py:class:`TimeAxis`.
    """
    TIME_CENTROID_EXTRA_PRECISION: Optional[Dataof[TimeSamplingArray]] = None
    """Additional precision for ``TIME_CENTROID``"""
    EFFECTIVE_CHANNEL_WIDTH: Optional[Dataof[FreqSamplingArray]] = None
    """The channel bandwidth that includes the effects of missing data."""
    FREQUENCY_CENTROID: Optional[Dataof[FreqSamplingArray]] = None
    """Includes the effects of missing data unlike ``frequency``."""

    # Optional Attributes
    pointing_xds: Optional[Attr[PointingXds]] = None
    source_xds: Optional[Attr[SourceXds]] = None
    pased_array_xds: Optional[Attr[PhasedArrayXds]] = None
    observation_info: Optional[Attr[ObservationInfoDict]] = None


@dataclass(frozen=True)
class PointingXds(AsDataset):
    # TODO
    pass

    
@dataclass(frozen=True)
class SpectralCoordXds(AsDataset):
    # TODO
    pass


@dataclass(frozen=True)
class AntennaXds(AsDataset):
    # TODO
    pass


@dataclass(frozen=True)
class SourceXds(AsDataset):
    # TODO
    pass


@dataclass(frozen=True)
class PhasedArrayXds(AsDataset):
    # TODO
    pass


@dataclass(frozen=True)
class DopplerXds(AsDataset):
    # TODO
    pass



import numpy as np
import xarray as xr

from .msv2_to_msv4_meta import column_description_casacore_to_msv4_measure
from .subtables import subt_rename_ids
from ._tables.read import read_generic_table


def create_ant_xds(in_file: str):
    """Creates an Antenna Xarray Dataset from a MS v2 ANTENNA table.

    Parameters
    ----------
    in_file : str
        Input MS name.

    Returns
    -------
    ant_xds : xarray.Dataset
        Antenna Xarray Dataset.
    """
    # Dictionaries that define the conversion from MSv2 to MSv4:
    to_new_data_variable_names = {
        "position": "POSITION",
        "offset": "FEED_OFFSET",
        "dish_diameter": "DISH_DIAMETER",
    }
    data_variable_dims = {
        "position": ["antenna_id", "xyz_label"],
        "offset": ["antenna_id", "xyz_label"],
        "dish_diameter": ["antenna_id"],
    }
    to_new_coord_names = {
        "name": "name",
        "station": "station",
        "type": "type",
        "mount": "mount",
        "phased_array_id": "phased_array_id",
    }
    coord_dims = {
        "name": ["antenna_id"],
        "station": ["antenna_id"],
        "type": ["antenna_id"],
        "mount": ["antenna_id"],
        "phased_array_id": ["antenna_id"],
    }

    # Read ANTENNA table into a Xarray Dataset.
    generic_ant_xds = read_generic_table(
        in_file,
        "ANTENNA",
        rename_ids=subt_rename_ids["ANTENNA"],
    )

    ant_column_description = generic_ant_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]
    # ['OFFSET', 'POSITION', 'TYPE', 'DISH_DIAMETER', 'FLAG_ROW', 'MOUNT', 'NAME', 'STATION']
    ant_xds = xr.Dataset()

    coords = {
        "antenna_id": np.arange(generic_ant_xds.sizes["antenna_id"]),
        "xyz_label": ["x", "y", "z"],
    }
    for key in generic_ant_xds:
        msv4_measure = column_description_casacore_to_msv4_measure(
            ant_column_description[key.upper()]
        )
        if key in to_new_data_variable_names:
            ant_xds[to_new_data_variable_names[key]] = xr.DataArray(
                generic_ant_xds[key].data, dims=data_variable_dims[key]
            )

            if msv4_measure:
                ant_xds[to_new_data_variable_names[key]].attrs.update(msv4_measure)

            if key in ["dish_diameter"]:
                ant_xds[to_new_data_variable_names[key]].attrs.update(
                    {"units": ["m"], "type": "quantity"}
                )

        if key in to_new_coord_names:
            coords[to_new_coord_names[key]] = (
                coord_dims[key],
                generic_ant_xds[key].data,
            )

    ant_xds = ant_xds.assign_coords(coords)
    return ant_xds


def create_weather_xds(in_file: str):
    """Creates a Weather Xarray Dataset from a MS v2 WEATHER table.

    Parameters
    ----------
    in_file : str
        Input MS name.

    Returns
    -------
    weather_xds : xarray.Dataset
        Weather Xarray Dataset.
    """
    # Dictionaries that define the conversion from MSv2 to MSv4:
    to_new_data_variable_names = {
        "h2o": "H2O",
        "ionos_electron": "IONOS_ELECTRON",
        "pressure": "PRESSURE",
        "rel_humidity": "REL_HUMIDITY",
        "temperature": "TEMPERATURE",
        "dew_point": "DEW_POINT",
        "wind_direction": "WIND_DIRECTION",
        "wind_speed": "WIND_SPEED",
    }
    data_variable_dims = {
        "h2o": ["station_id", "time"],
        "ionos_electron": ["station_id", "time"],
        "pressure": ["station_id", "time"],
        "rel_humidity": ["station_id", "time"],
        "temperature": ["station_id", "time"],
        "dew_point": ["station_id", "time"],
        "wind_direction": ["station_id", "time"],
        "wind_speed": ["station_id", "time"],
    }
    to_new_coord_names = {
        # No MS data cols are turned into xds coords
    }
    coord_dims = {
        # No MS data cols are turned into xds coords
    }
    to_new_dim_names = {
        "antenna_id": "station_id",
    }

    # Read WEATHER table into a Xarray Dataset.
    try:
        generic_weather_xds = read_generic_table(
            in_file,
            "WEATHER",
            rename_ids=subt_rename_ids["WEATHER"],
        )
    except ValueError as _exc:
        return None

    generic_weather_xds = generic_weather_xds.rename_dims(to_new_dim_names)

    weather_column_description = generic_weather_xds.attrs["other"]["msv2"][
        "ctds_attrs"
    ]["column_descriptions"]
    # ['ANTENNA_ID', 'TIME', 'INTERVAL', 'H2O', 'IONOS_ELECTRON',
    #  'PRESSURE', 'REL_HUMIDITY', 'TEMPERATURE', 'DEW_POINT',
    #  'WIND_DIRECTION', 'WIND_SPEED']
    weather_xds = xr.Dataset()

    coords = {
        "station_id": generic_weather_xds["station_id"].data,
        "time": generic_weather_xds["time"].data,
    }
    for key in generic_weather_xds:
        msv4_measure = column_description_casacore_to_msv4_measure(
            weather_column_description[key.upper()]
        )
        if key in to_new_data_variable_names:
            var_name = to_new_data_variable_names[key]
            weather_xds[var_name] = xr.DataArray(
                generic_weather_xds[key].data, dims=data_variable_dims[key]
            )

            if msv4_measure:
                weather_xds[var_name].attrs.update(msv4_measure)

            if key in ["interval"]:
                weather_xds[var_name].attrs.update({"units": ["s"], "type": "quantity"})
            elif key in ["h2o"]:
                weather_xds[var_name].attrs.update(
                    {"units": ["/m^2"], "type": "quantity"}
                )
            elif key in ["ionos_electron"]:
                weather_xds[var_name].attrs.update(
                    {"units": ["/m^2"], "type": "quantity"}
                )
            elif key in ["pressure"]:
                weather_xds[var_name].attrs.update(
                    {"units": ["Pa"], "type": "quantity"}
                )
            elif key in ["rel_humidity"]:
                weather_xds[var_name].attrs.update({"units": ["%"], "type": "quantity"})
            elif key in ["temperature"]:
                weather_xds[var_name].attrs.update({"units": ["K"], "type": "quantity"})
            elif key in ["dew_point"]:
                weather_xds[var_name].attrs.update({"units": ["K"], "type": "quantity"})
            elif key in ["wind_direction"]:
                weather_xds[var_name].attrs.update(
                    {"units": ["rad"], "type": "quantity"}
                )
            elif key in ["wind_speed"]:
                weather_xds[var_name].attrs.update(
                    {"units": ["m/s"], "type": "quantity"}
                )

        if key in to_new_coord_names:
            coords[to_new_coord_names[key]] = (
                coord_dims[key],
                generic_weather_xds[key].data,
            )

    weather_xds = weather_xds.assign_coords(coords)
    return weather_xds


def create_pointing_xds(in_file: str):
    """Creates a Pointing Xarray Dataset from an MS v2 POINTING (sub)table.

    WIP: details of a few direction variables (and possibly moving some to attributes) to be
    settled (see MSv4 spreadsheet).

    Parameters
    ----------
    in_file : str
        Input MS name.

    Returns
    -------
    pointing_xds : xarray.Dataset
        Pointing Xarray Dataset.
    """

    # Dictionaries that define the conversion from MSv2 to MSv4:
    to_new_data_variable_names = {
        # "name": "NAME",   # removed
        # "time_origin": "TIME_ORIGIN",  # removed?
        "direction": "BEAM_POINTING",
        "encoder": "DISH_MEASURED_POINTING",
        "target": "TARGET",  # => attribute?
        "pointing_offset": "POINTING_OFFSET",
        "source_offset": "SOURCE_OFFSET",
        # "pointing_model_id": "POINTING_MODEL_ID",   # removed
        # "tracking": "TRACKING",   # => attribute
        # "on_source": "ON_SOURCE",   # removed
        "over_the_top": "OVER_THE_TOP",
    }
    time_ant_ids = ["time", "antenna_id"]
    data_variable_dims = {
        # "name": ["time", "antenna_id"],   # removed
        # "time_origin": ["time", "antenna_id"],   # removed?
        "direction": ["time", "antenna_id", "direction"],
        "encoder": ["time", "antenna_id", "direction"],
        "target": ["time", "antenna_id", "direction"],
        "pointing_offset": ["time", "antenna_id", "direction"],
        "source_offset": ["time", "antenna_id", "direction"],
        # "pointing_model_id": ["time", "antenna_id"],   # removed
        # "tracking": ["time", "antenna_id"],   # => attribute
        # "on_source": ["time", "antenna_id"],  # removed
        "over_the_top": ["time", "antenna_id"],
    }
    to_new_coord_names = {"ra/dec": "direction"}
    coord_dims = {}

    # Read POINTING table into a Xarray Dataset.
    generic_pointing_xds = read_generic_table(
        in_file,
        "POINTING",
        rename_ids=subt_rename_ids["POINTING"],
    )
    if not generic_pointing_xds.data_vars:
        # apparently empty MS/POINTING table => produce empty xds
        return xr.Dataset()

    # Checking a simple way of using only the one single coefficient of the polynomials
    if "n_polynomial" in generic_pointing_xds.sizes:
        size = generic_pointing_xds.sizes["n_polynomial"]
        if size == 1:
            generic_pointing_xds = generic_pointing_xds.sel({"n_polynomial": 0})

    pointing_column_descriptions = generic_pointing_xds.attrs["other"]["msv2"][
        "ctds_attrs"
    ]["column_descriptions"]

    pointing_xds = xr.Dataset()
    for key in generic_pointing_xds:
        if key in to_new_data_variable_names:
            data_var_name = to_new_data_variable_names[key]
            pointing_xds[data_var_name] = xr.DataArray(
                generic_pointing_xds[key].data, dims=data_variable_dims[key]
            )

            msv4_measure = column_description_casacore_to_msv4_measure(
                pointing_column_descriptions[key.upper()]
            )
            if msv4_measure:
                pointing_xds[data_var_name].attrs.update(msv4_measure)

    coords = {
        "time": generic_pointing_xds["time"].values,
        "antenna_id": np.arange(generic_pointing_xds.sizes["antenna_id"]),
        "direction": ["ra", "dec"],
    }
    pointing_xds = pointing_xds.assign_coords(coords)

    # missing attributes
    pointing_xds["time"].attrs.update({"units": ["s"], "type": "quantity"})

    if "tracking" in generic_pointing_xds.data_vars:
        pointing_xds.attrs["tracking"] = generic_pointing_xds.data_vars[
            "tracking"
        ].values[0, 0]

    # Move target from data_vars to attributes?
    move_target_as_attr = False
    if move_target_as_attr:
        target = generic_pointing_xds.data_vars["target"]
        pointing_xds.attrs["target"] = {
            "dims": ["direction"],
            "data": target.values[0, 0].tolist(),
            "attrs": column_description_casacore_to_msv4_measure(
                pointing_column_descriptions["TARGET"]
            ),
        }
    # TODO: move also source_offset/pointing_offset from data_vars to attrs?

    return pointing_xds



from .metamodel import AttrSchemaRef, ArraySchema, ArraySchemaRef, DatasetSchema
from .dataclass import (
    xarray_dataclass_to_array_schema,
    xarray_dataclass_to_dataset_schema,
)
from .bases import AsDataArray, AsDataset

__all__ = [
    "AttrSchemaRef",
    "ArraySchema",
    "ArraySchemaRef",
    "DatasetSchema",
    "xarray_dataclass_to_array_schema",
    "xarray_dataclass_to_dataset_schema",
    "AsDataArray",
    "AsDataset",
]
